# Generated by ariadne-codegen
# Source: operations.graphql

from stigg._vendors.pydantic import Field

from .base_model import BaseModel


class CancelSubscriptionUpdates(BaseModel):
    cancel_schedule: str = Field(alias="cancelSchedule")
